#!/usr/local/bin/python
#
# a python class to interface with freebsd's userppp package.
#
# 0.8.1, Fri Mar 19 15:24:06 1999
# johnchia@innocent.com
#

from string import split,join,find,lower
from time import sleep
from telnetlib import *
from socket import error

class pppctlClass:

	"PPP Control Class for FreeBSD's User-space PPP package."

	def __init__(self, host=0, port=0, passwd=0, user=0):
		self.my_buf = ""
		self.err_msg = ""
		self.in_buf = ""

		self.timeout = 4		# default timeout for recv'ing (in seconds)

		self.pollrate = 0.03		# how many seconds to wait between each poll
						# used when self.return_when_connected is true

		self.return_when_complete = 1	# return only when completely connected, or
						# completely down, set to 0 for returning right 
						# after the command is sent (start_<_0.8-esque)

		self.lprompt = ""		# lower-cased prompt

		if host and port and passwd:
			self.open_ppp( host, port, passwd, user )




	def open_ppp ( self, host, port, passwd , prompt = "> ", user = 0):
		""" 	open_ppp - Connect and authenticate
			returns:
				-1 error connecting
			 	0 bad password, closed
			 	1 open, authenticated, ready.	"""
		try:
			self.ppp = Telnet(host, port)		# open host
		except socket.error:
			return(-1)				# exit if error

		self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt
		self.lprompt = lower(self.my_buf)
		self.ppp.write('passwd '+passwd)		# send password

		self.my_buf = self.ppp.read_until(prompt, self.timeout)	# read until prompt
		if find(self.my_buf, "ON") == -1: return(0)		# make sure password went through OK

		return(1)


	# --
	# send_ppp - send data out the pipe
	# returns what was sent back minus prompt

	def send_ppp ( self, data, prompt = "> " ):
		self.in_buf = self.ppp.read_very_eager()

		self.ppp.write( data+"\n" )

		self.in_buf = ['']
		while self.in_buf == [''] or lower(self.in_buf) == self.lprompt:
			self.in_buf = self.ppp.read_until(prompt,self.timeout)

		self.in_buf = split(self.in_buf,"\n")
		self.in_buf.remove(self.in_buf[len(self.in_buf)-1])	# remove prompt
		self.in_buf = join(self.in_buf,"\n")

		return(self.in_buf)


	# --
	# Throughput Function - get one link's throughput
	# returns overall, peak, date of peak, link uptime (seconds)

	def throughput(self, conn):
	        my_buf = self.send_ppp ( "link "+conn+" show modem" )                       
		my_buf = split(my_buf,'\n')[-5:]	# split by \n and take last 5 lines

	        overall	= split( my_buf[2] ) [1]	# average
	        peak 	= split( my_buf[4] ) [1]	# peak
	        peak_on	= join( split( my_buf[4] ) [4:] )	# date of peak
		link_up	= split( my_buf[0] ) [2]	# link uptime in seconds

		return( overall, peak, peak_on, int(link_up) )


	# --
	# status - get all links and their status
	# returns list of lists, each list in list is a defined connection..
	# 	0 <name> 
	#	1 <open/dial/closed/etc)
	#	2 [current throughput in bytes]
	#	3 [overall throughput] 
	#	4 [peak throughput] 
	#	5 [date of peak]
	#	6 [connect time]
	#	-- indexes 2-6 are returned only when connection is open.

	def status(self):

		out = []
		self.my_buf = split( self.send_ppp('show links'), "\n" )

	        for x in range( len( self.my_buf ) ):
			out.append([])
			cur_link = split(self.my_buf[x])	# split current link (one per line)

			out[x].append( cur_link[1] )		# put name -> out
			out[x].append( cur_link[ 3 ][ :-1 ] )	# put status -> out

			if cur_link[3][0:-1] == "open":		# open, so get throughput too
				out[x].append( cur_link[6] )	# current
				for a in range(4): out[x].append("")
				out[x][3], out[x][4], out[x][5], out[x][6] = self.throughput( cur_link[1] )	# Get throughput
			else:					# append zeros
				for a in range(5): out[x].append(0)
		return(out)


	# --
	# get_status - get one link's status
	# returns status (open,dial,closed,etc) of conn

	def get_status(self,conn):

		# This is `kind-of-messy' but it works. (it doesn't without it)
		# Someone (me) `should' find out why this is
		self.my_buf = ['']
		while self.my_buf == ['']:
			self.my_buf = split( self.send_ppp('show links'), "\n" )

	        for x in range( len( self.my_buf ) ):

			cur_link = split(self.my_buf[x])		# split current link (one per line)

			if cur_link[1] == conn:
				if cur_link[3][:-1] == 	"closed" 	or \
							"opening"	or \
							"hangup" 	or \
							"dial"		or \
							"login"		or \
							"ready"		or \
							"lcp" 		or \
							"auth" 		or \
							"cbcp" 		or \
							"open":
					return( cur_link[3][:-1] )
				else:
					self.err_msg = "ERROR: Unknown state: "+cur_link[3][:-1]
					return(0)		# error
		self.err_msg = "ERROR: couldn't find connection"
		return(0)

	# --
	# dial_ppp - dial a link
	# returns 1 for success, 0 for failure, error mesg stored in self.err_msg

	def dial_ppp ( self, conn ):
	
		self.my_buf = self.send_ppp ( 'link ' + conn + ' dial' )	# send command

		if self.my_buf != "":				# there was some sort of error.
			self.err_msg = self.my_buf
			return(0)

		if not self.return_when_complete: return(1)	# return right away if that's
								# what we're supposed ta do

		sleep(self.pollrate)					# sleep a second to make sure
		status = "null"
		while status != "open" and status != "err" and status != "closed":
			sleep(self.pollrate)
			status = self.get_status(conn)
		if status == "open": return(1)
		else: return(0)


	# --
	# down_ppp - down a link
	# Returns 1 for success

	def down_ppp( self, conn ):
		self.my_buf = self.send_ppp ( 'link ' + conn + ' down' )		# send command

		if self.my_buf != "":					# there was some sort of error.
			self.err_msg = self.my_buf
			return(1)

		if not self.return_when_complete: return(0)		# return right away if that's
									# what we're supposed ta do

		sleep(self.pollrate)		# sleep a second to make sure we've begun dialing
		status = "null"
		while status != "open" and status != "err" and status != "closed":
			sleep(self.pollrate)
			status = self.get_status(conn)
		if status == "closed": return(1)
		else: return(0)


	# --
	# close_ppp - closes connection
	# disconnects from the PPP control server

	def close_ppp(self):
		self.ppp.close()


"""
#
# Connection List
# replaced with better status function
# returns a list of connections
#
def conn_list():
	conns = s_r("show links\n")

	conns.reverse()
	conns.remove(conns[0])
	conns.reverse()
	
	for x in range(len(conns)):
		y = conns[x]
		y = self.split(y)
		conns[x] = y[1]

	return(conns)

States (from /usr/src/usr.sbin/ppp/datalink.c freebsd-STABLE 3.1-199902XX)
  "closed",
  "opening",
  "hangup",
  "dial",
  "login",
  "ready",
  "lcp",
  "auth",
  "cbcp",
  "open"

"""

